from generators import *
import tools


class ClausesDataset:
    '''
    Storage of clauses; use .clauses to get list of clauses (:type: Clause).
    
    Use .xFactory to get predicate, variable, constant, functor factory.
    '''

    def __init__(self, clauses, predicateFactory, variableFactory, constantFactory, functorFactory):
        self.clauses = list(clauses)
        self.predicateFactory = predicateFactory
        self.variableFactory = variableFactory
        self.constantFactory = constantFactory
        self.functorFactory = functorFactory


class ClausesDatasetBuilder:
    def __init__(self, predicateFactory=PredicateFactory(),variableFactory=VariableFactory()):
        '''
        Creates stateful builder with predicateFactory, constantFactory, functorFactory and variableFactory.
        '''
        self.clauses = []
        self.predicateFactory = predicateFactory
        self.constantFactory = ConstantFactory()
        self.variableFactory = variableFactory
        self.functorFactory = FunctorFactory()

    def append(self, clause):
        '''
        Append another clause.

        :type clause: Clause
        :rtype: ClausesDatasetBuilder
        '''
        self.clauses.append(clause)
        return self

    def build(self):
        '''
        Creates and returns ClausesDataset from collected samples.

        :rtype: Dataset
        '''
        return ClausesDataset(self.clauses,self.predicateFactory,self.variableFactory,self.constantFactory,self.functorFactory)



class Dataset:
    '''
    Storage of samples. use .samples to get list of samples (:type: Sample).
    '''

    def __init__(self, samples):
        self.samples = samples

    def getPredicates(self):
        '''
        Returns set of all predicates used in the dataset.

        :rtype: set(Predicate)
        '''
        literals = map(lambda sample: sample.getPredicates(), self.samples)
        return tools.unionSets(literals)

    def getConstants(self):
        '''
        Returns all constants in the dataset.

        :rtype: set(Constant)
        '''
        constants = map(lambda sample: sample.getConstants(), self.samples)
        return tools.unionSets(constants)

    def maxArity(self):
        '''
        Returns maximal arity among the predicates in dataset.

        :rtype: int
        '''
        return max(map(lambda predicate: predicate.arity, self.getPredicates()))


class DatasetBuilder:
    def __init__(self, predicateFactory=PredicateFactory()):
        '''
        Creates stateful dataset of interpretations builder with predicateFactory, constantFactory, functorFactory and variableFactory.
        '''
        self.samples = []
        self.predicateFactory = predicateFactory
        self.constantFactory = ConstantFactory()
        self.variableFactory = VariableFactory()
        self.functorFactory = FunctorFactory()

    def append(self, sample):
        '''
        Append another sample.

        :type sample: Sample
        :rtype: DatasetBuilder
        '''
        self.samples.append(sample)
        return self

    def build(self):
        '''
        Creates and returns dataset from collected samples.

        :rtype: Dataset
        '''
        return Dataset(self.samples)


class Sample:
    '''
    Stores a sample which consist of a class label and an interpretation.

    Class is either positive '+' or negative '-'.
    Delimiter ':' splits class and the interpretation which is ended by a dot.

    use .positiveClass to get T if the sample has positive label \n
    use .interpretation to get interpretation stored in the sample (:type: Interpretation)
    '''

    def __init__(self, interpretation, positiveClass):
        self.interpretation = interpretation
        self.positiveClass = positiveClass

    def __str__(self):
        sign = "+"
        if not self.positiveClass:
            sign = "-"
        return sign + " : " + str(self.interpretation)

    def getPredicates(self):
        '''
        Returns set of predicates.

        :rtype: set(Predicate)
        '''
        return self.interpretation.getPredicates()

    def getConstants(self):
        '''
        Returns set of constants.

        :rtype: set(Constant)
        '''
        return self.interpretation.getConstants()
