from logic import *


class ConstantFactory:
    def __init__(self):
        self.memory = {}

    def create(self, name):
        '''
        Returns constant based on the given name.

        :type name: str
        :rtype: Constant
        '''
        if len(name) < 1:
            raise Exception("The name must be non-empty.")
        elif name[0].isupper():
            raise Exception("The name must start with a lowercase; '"+name+"' given instead.")

        if name not in self.memory.keys():
            self.memory[name] = Constant(name)
        return self.memory[name]


class FunctorFactory:
    def __init__(self):
        self.memory = {}

    def create(self, name,arity):
        '''
        Returns functor based on the given name and arity.

        :type name: str
        :rtype: Functor
        '''
        if len(name) < 1:
            raise Exception("The name must be non-empty.")
        elif name[0].isupper():
            raise Exception("The name must start with a lowercase.")

        if (name,arity) not in self.memory.keys():
            self.memory[(name,arity)] = Functor(name, arity)
        return self.memory[(name,arity)]


class VariableFactory:
    def __init__(self):
        self.memory = []

    def create(self):
        '''
        Returns new variable.

        :rtype: Variable
        '''
        count = len(self.memory)
        variable = Variable("X" + str(count))
        self.memory.append(variable)
        return variable


class PredicateFactory:
    def __init__(self):
        self.memory = {}

    def create(self, name, arity):
        '''
        Returns predicate based on the given name and arity

        :type name: string
        :type arity: int,>=0
        :rtype: Predicate
        '''

        if len(name) < 1:
            raise Exception("Predicate name must be non-empty.")
        elif name[0].isupper():
            raise Exception("Predicate name must start with a lowercase.")

        if (name, arity) not in self.memory.keys():
            self.memory[(name, arity)] = Predicate(name, arity)
        return self.memory[(name, arity)]


class DataBuilder:
    def __init__(self):
        pass

    def append(self):
        pass

    def build(self):
        pass


class ClauseBuilder(DataBuilder):
    def __init__(self):
        self.literals = []

    def append(self, literal):
        '''
        Adds literal to the clause.

        :type literal: Literal
        :rtype: ClauseBuilder
        '''
        self.literals.append(literal)
        return self

    def build(self):
        ''' 
        Creates and returns Clause from the builder.
        
        :rtype: Clause
        '''
        return Clause(self.literals)


class InterpretationBuilder(DataBuilder):
    '''
    Builder for interpretations.
    '''

    def __init__(self):
        self.literals = []

    def append(self, literal):
        '''
        Adds literal to the builder.

        :type literal: Literal
        :rtype: InterpretationBuilder
        '''
        self.literals.append(literal)
        return self

    def build(self):
        '''
        Creates and returns an interpretation from the builder.

        :rtype: Interpretation
        '''
        return Interpretation(self.literals, None)


class MostSpecificCNFGenerator:
    '''
    Generator of most specific CNF which has to be firstly constructed, given some settings, and then run by .generated().

    Note that functors with arity greater than 0 are not supported, so the flag is there for future usage.

    Note that the options useConstant is on you to be implemented if you want to.
    '''

    def __init__(self, predicates, s, t, constants=[], functors=[]):
        '''
        Creates new generator, which firstly stores the setting, e.g. set of predicates which should be used, and s and t parameter for generating all possible st-range restricted clauses.

        St-clause is such a clause which has at most s literals, each of them is composed by at most t terms and predicate symbols.

        Range-restricted clause is such a clause that all of its variables in positive literals occur in some negative literals.

        :type predicates: set of Predicate
        :type s: int,>=0
        :type t: int,>=0
        :type constants: set of Constant
        :type functors: set of Functor
        :rtype: MostSpecificCNFGenerator
        '''
        # corresponding naming from the lecture, use generate() method to generate the most specific
        # current version does not use functors symbols
        self.predicates = predicates
        self.s = s
        self.t = t
        self.constants = constants
        self.functors = functors
        self.variableFactory = VariableFactory()

    def filterRedundant(self, clauses):
        cannonic = []
        canonics = set()
        for clause in filter(lambda clause: not clause.hasRedundantLiteral(), clauses):
            canonic = clause.getCanonicString()
            if canonic not in canonics:
                cannonic.append(clause)
                canonics.add(canonic)
        return cannonic

    def generate(self, useConstants=False):
        '''
        Returns most specific FOL CNF given predicates and constrains s and t.

        Note that the options useConstant is on you to be implemented if you want to, but the labs can be done without that. Implementing the class GeneralizatingAgent is more important.

        If useConstant is set to True, then all possible combinations given constants are generated. If the parameter is set to False, only clauses with variables are generated.

        :type useConstants: bool
        :rtype: CNF
        '''
        st = self.generateST()
        if not st:
            return CNF(st, self.variableFactory)
        restricted = self.generateRangeRestricted(st)
        fresh = list(self.generateFreshVariableForEachClause(restricted))

        if useConstants:
            fresh.extend(self.generateWithConstants(fresh))

        fresh.append(Clause([]))
        return CNF(fresh, self.variableFactory)

    def generateNext(self, previous):
        generated = set()
        for predicate in self.predicates:
            for clause in previous:
                for combination in self.generateTermsExtensions(clause.getVariables(), predicate):
                    atom = Atom(predicate, combination)
                    generated.add(clause.extend(Literal(atom, False)))

        return generated

    def generateAllCombinations(self, predicates):
        '''
        Generates set of clauses which can be generated from the predicates, given s and t constrains

        :type predicates: set of Predicates
        :rtype: set of Clause
        '''
        oneLiteralClauses = set()
        for predicate in filter(lambda predicate: predicate.arity < self.t, predicates):
            for term in self.generateTermsExtensions([], predicate):
                atom = Atom(predicate, term)
                oneLiteralClauses.add(Literal(atom, False))
        clauses = list(map(lambda literal: Clause([literal]), oneLiteralClauses))
        return clauses

    def generateTermsExtensions(self, usedVariables, predicate):
        def resolvePosition(usedSymbols, variables, functions, missingArguments):
            if missingArguments < 1:
                return [[]]
            successorsVariables = set(variables)
            successorsVariables.add(self.variableFactory.create())
            result = []
            for variable in successorsVariables:
                for successor in resolvePosition(usedSymbols + 1, successorsVariables, functions, missingArguments - 1):
                    term = [variable]
                    term.extend(successor)
                    result.append(term)
            return result

        return resolvePosition(1, usedVariables, self.functors, predicate.arity)

    def generateFreshVariableForEachClause(self, clauses):
        return map(lambda clause: tools.replaceWithFreshVariables(clause, self.variableFactory), clauses)

    def generateST(self):
        previous = self.filterRedundant(self.generateAllCombinations(self.predicates))
        generated = list(previous)
        for loop in range(1, self.s):
            possibleCombinations = self.generateNext(previous)
            previous = self.filterRedundant(possibleCombinations)
            generated.extend(previous)
        return generated

    def generateRangeRestricted(self, generated):
        equal = (lambda clauses: filter(lambda clause: clause.getNumberOfLiterals() == self.s, clauses))
        undone = (lambda clauses: list(filter(lambda clause: clause.getNumberOfLiterals() < self.s, clauses)))
        result = set(generated)
        nonSaturated = undone(generated)
        for idx in range(1, self.s):
            current = self.addPositiveRestrictedLiterals(nonSaturated)
            result.update(equal(current))
            nonSaturated = undone(current)
            result.update(nonSaturated)

        return result

    def addPositiveRestrictedLiterals(self, nonSaturated):
        return tools.unionSets(map(lambda clause: self.extendByHeads(clause), nonSaturated))

    def extendByHeads(self, clause):
        result = set()
        variables = list(clause.getNegativeLiteralsAsClause().getVariables())
        atoms = list(map(lambda literal: str(literal.atom), clause.literals))
        for predicate in filter(lambda predicate: predicate.arity < self.t, self.predicates):
            for permutation in itertools.permutations(variables, predicate.arity):
                head = Literal(Atom(predicate, permutation), True)
                if str(head.atom) not in atoms:
                    result.add(clause.extend(head))

        return result

    def generateWithConstants(self, clauses):
        '''
        You may either use all the st-range restricted lifted-only clauses given (all terms are variables) or implement the method from scratch.

        The method should generate all possible st-range restricted clauses.

        :type clauses: List(Clause)
        :rtype: List(Clause)
        '''
        ### your code here
        pass
        return set()
