import os
from dataset import *


class Parser:
    '''
    Parser for the dataset (list of interpretations, each one loaded in Sample class).

    Use the method parseDatasetOfInterpretations() to get dataset of interpretations or parseClauses() to get dataset of clauses.

    '''

    def __init__(self, pathToDataFile,variableFactory=VariableFactory(),predicateFactory=PredicateFactory()):
        '''
        Constructs new parser. Use parse() method to get dataset.

        :type pathToDataFile: str
        :type variableFactory: VariableFactory
        :type predicateFactory: PredicateFactory
        :rtype: Parser
        '''
        self.path = pathToDataFile
        self.CLASS_CLAUSE_DELIMITER = ':'
        self.ENDING_CHARACTER = '.'
        self.NEGATION_SIGN = '-'
        self.OPENING_BRACKET = '('
        self.CLOSING_BRACKET = ')'
        self.SEPARATOR = ','
        self.COMMENT_LINE = '#'
        self.predicateFactory = predicateFactory
        self.variableFactory = variableFactory

    def check(self):
        if not os.path.isfile(self.path):
            raise Exception('The file was not found \'' + self.path + '\'.')

    def notParseable(self, line):
        return len(line.strip()) < 1 or line.strip().startswith(self.COMMENT_LINE)

    def parseClauses(self):
        '''
        Parses the file as a list of clauses and returns them ClausesDataset.
        
        :rtype: ClausesDataset 
        '''
        self.check()

        builder = ClausesDatasetBuilder(predicateFactory=self.predicateFactory,variableFactory=self.variableFactory)
        with open(self.path, 'r') as file:
            for line in file:
                if self.notParseable(line):
                    continue
                clause = self.parseLine(line, builder.predicateFactory, builder.constantFactory,
                                        builder.variableFactory, builder.functorFactory, ClauseBuilder())
                builder.append(clause)

        return builder.build()

    def parseDatasetOfInterpretations(self):
        '''
        Parses the given file to Dataset of interpretations.

        Parses sample in format 'class : parent(johan,anna),female(marie),male(johan).'.
        Class is either positive '+' or negative '-'. Delimiter ':' splits class and the interpretation which is ended by a dot.

        Possible exception if the file was not found.

        :rtype: Dataset
        '''
        self.check()

        builder = DatasetBuilder(predicateFactory=self.predicateFactory)
        with open(self.path, 'r') as file:
            for line in file:
                if self.notParseable(line):
                    continue
                classPart = line[:line.find(self.CLASS_CLAUSE_DELIMITER)].strip()
                positiveClass = True
                if "-" == classPart:
                    positiveClass = False
                if classPart not in ["+", "-"]:
                    raise Exception("Unknown class token used '" + classPart + "'.")
                interpretationPart = line[line.find(self.CLASS_CLAUSE_DELIMITER) + 1:].strip()
                cnf = self.parseLine(interpretationPart, builder.predicateFactory, builder.constantFactory,
                                     builder.variableFactory, builder.functorFactory, InterpretationBuilder())
                builder.append(Sample(cnf, positiveClass))

        return builder.build()

    def variableMapping(self, mapping, text, variableFactory):
        if text not in mapping:
            mapping[text] = variableFactory.create()
        return mapping[text]

    def retrieveArguments(self, line, predicateFactory, constantFactory, variableFactory, functorFactory,
                          textToVariable={}):
        if len(line) < 1:
            raise Exception("Line ended too shortly ('" + line + "').")
        assert self.OPENING_BRACKET == line[0]

        arguments = []
        line = line[1:]
        while True:
            assert len(line) > 0
            if line.startswith(self.CLOSING_BRACKET):
                line = line[1:]
                break

            comma = line.find(self.SEPARATOR)
            opening = line.find(self.OPENING_BRACKET)
            closing = line.find(self.CLOSING_BRACKET)

            if (opening < comma and -1 != opening) or (-1 != opening and opening < closing):
                # parsing function
                name = line[:opening]
                functionArguments, line = self.retrieveArguments(line[opening:], predicateFactory, constantFactory,
                                                                 variableFactory, functorFactory,
                                                                 textToVariable=textToVariable)
                functor = functorFactory.create(name, len(functionArguments))
                arguments.append(CompoundTerm(functor, functionArguments))
            else:
                # parsing variable or constant
                if -1 != closing and (closing < comma or comma < 0):
                    name = line[:closing]
                    line = line[closing:]
                else:
                    name = line[:comma]
                    line = line[comma:]
                assert len(name) > 0
                if name[0].islower() or name.replace('.', '', 1).isdigit():
                    arguments.append(constantFactory.create(name))
                else:
                    arguments.append(self.variableMapping(textToVariable, name, variableFactory))
            if line.startswith(self.SEPARATOR):
                line = line[1:]
        if line.startswith(self.SEPARATOR):
            line = line[1:]

        return arguments, line

    def parseLine(self, line, predicateFactory, constantFactory, variableFactory, functorFactory, builder):
        '''
        Parses the line according to builder (either InterpretationBuilder or ClauseBuilder).  
        
        :type line: str
        :type predicateFactory: PredicateFactory
        :type constantFactory: ConstantFactory
        :type variableFactory: VariableFactory
        :type functorFactory: FunctorFactory
        :type builder: DataBuilder
        :rtype: tuple of class (T|F) and parsed clause
        '''

        while True:
            line = line.strip()
            if self.ENDING_CHARACTER == line[0]:
                break
            positive = True
            if self.NEGATION_SIGN == line[0]:
                positive = False
                line = line[1:]

            comma = line.find(self.SEPARATOR)
            opening = line.find(self.OPENING_BRACKET)
            endingDot = line.find(self.ENDING_CHARACTER)

            if isinstance(builder, InterpretationBuilder):
                assert True == positive

            if (comma < opening and -1 != comma): # or (-1 == opening and -1 == comma):
                # predAr0,anotherPred(arg1,arg2)...
                predicateName = line[:comma]
                atom = Atom(predicateFactory.create(predicateName, 0), [])
                builder.append(Literal(atom, positive))
                line = line[comma + 1:].strip()
            elif endingDot < opening or (-1 == opening):
                # lastPredicateWithArityZero.
                predicateName = line[:endingDot]
                atom = Atom(predicateFactory.create(predicateName, 0), [])
                builder.append(Literal(atom, positive))
                break
            else:
                predicateName = line[:opening]
                arguments, line = self.retrieveArguments(line[opening:], predicateFactory, constantFactory,
                                                         variableFactory, functorFactory)
                atom = Atom(predicateFactory.create(predicateName, len(arguments)), arguments)
                builder.append(Literal(atom, positive))
        return builder.build()
